Namespace PredictionMarket
    Friend Class PredictionMarketProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbLf & "Begin prediction market demo " & vbLf)
            Console.WriteLine("Goal is to predict winner of Xrays (team [0]) ")
            Console.WriteLine("vs. Yanks (team [1]) using expert opinions ")

            Dim liq As Double = 100.0 ' liquidity
            Console.WriteLine(vbLf & "Setting liquidity parameter = " & liq.ToString("F1"))

            Dim outstanding() As Integer = {0, 0}
            Console.WriteLine(vbLf & "Initial number of shares owned of teams [0] and [1] are: ")
            ShowVector(outstanding)

            Dim probs() As Double = Probabilities(outstanding, liq)
            Console.WriteLine("Initial inferred probabilities of winning are: ")
            ShowVector(probs, 4, " ")

            Console.WriteLine("=================================")

            Dim costPerShare() As Double = CostForOneShare(outstanding, liq)
            Console.WriteLine("Current costs for one share of each team are: ")
            ShowVector(costPerShare, 4, " $")

            Console.WriteLine("Update: expert [01] buys 20 shares of team [0]")
            Dim costTrans As Double = CostOfTrans(outstanding, 0, 20, liq)
            Console.WriteLine("Cost of transaction to expert was: $" & costTrans.ToString("F2"))

            outstanding = New Integer() {20, 0}
            Console.WriteLine(vbLf & "New number of shares owned of teams [0] and [1] are: ")
            ShowVector(outstanding)

            probs = Probabilities(outstanding, liq)
            Console.WriteLine("New inferred probabilities of winning are: ")
            ShowVector(probs, 4, " ")

            Console.WriteLine("=================================")

            costPerShare = CostForOneShare(outstanding, liq)
            Console.WriteLine("Current costs for one share of each team are: ")
            ShowVector(costPerShare, 4, " $")

            Console.WriteLine("Update: expert [02] buys 20 shares of team [1]")
            costTrans = CostOfTrans(outstanding, 1, 20, liq)
            Console.WriteLine("Cost of transaction to expert was: $" & costTrans.ToString("F2"))

            outstanding = New Integer() {20, 20}
            Console.WriteLine(vbLf & "New number of shares owned of teams [0] and [1] are: ")
            ShowVector(outstanding)

            probs = Probabilities(outstanding, liq)
            Console.WriteLine("New inferred probabilities of winning are: ")
            ShowVector(probs, 4, " ")

            Console.WriteLine("=================================")

            costPerShare = CostForOneShare(outstanding, liq)
            Console.WriteLine("Current costs for one share of each team are: ")
            ShowVector(costPerShare, 4, " $")

            Console.WriteLine("Update: expert [03] buys 60 shares of team [0]")
            costTrans = CostOfTrans(outstanding, 0, 60, liq)
            Console.WriteLine("Cost of transaction to expert was: $" & costTrans.ToString("F2"))

            outstanding = New Integer() {80, 20}
            Console.WriteLine(vbLf & "New number of shares owned of teams [0] and [1] are: ")
            ShowVector(outstanding)

            probs = Probabilities(outstanding, liq)
            Console.WriteLine("New inferred probabilities of winning are: ")
            ShowVector(probs, 4, " ")

            Console.WriteLine("=================================")

            costPerShare = CostForOneShare(outstanding, liq)
            Console.WriteLine("Current costs for one share of each team are: ")
            ShowVector(costPerShare, 4, " $")

            Console.WriteLine("Update: expert [01] sells 10 shares of team [0]")
            costTrans = CostOfTrans(outstanding, 0, -10, liq)
            Console.WriteLine("Cost of transaction to expert was: $" & costTrans.ToString("F2"))

            outstanding = New Integer() {70, 20}
            Console.WriteLine(vbLf & "New number of shares owned of teams [0] and [1] are: ")
            ShowVector(outstanding)

            probs = Probabilities(outstanding, liq)
            Console.WriteLine("New inferred probabilities of winning are: ")
            ShowVector(probs, 4, " ")

            Console.WriteLine("=================================")

            Console.WriteLine("Update: Market Closed")
            Console.WriteLine(vbLf & "End prediction market demo " & vbLf)
            Console.ReadLine()

        End Sub ' Main()

        Private Shared Function Probabilities(ByVal outstanding() As Integer, ByVal liq As Double) As Double()
            ' aka marginal price
            ' return the inferred probabilities of winning
            ' from the outstanding shares owned on each team/option
            ' b = liquidity. smaller values = faster change
            Dim result(1) As Double
            Dim denom As Double = 0.0
            For i As Integer = 0 To 1
                denom += Math.Exp(outstanding(i) / liq)
            Next i
            For i As Integer = 0 To 1
                result(i) = Math.Exp(outstanding(i) / liq) / denom
            Next i
            Return result
        End Function ' Probs

        'Static Function OutstandingFromProbs(ByVal probs() As Double, ByVal liq As Double) As Integer()
        '    ' return a set of outstanding[] shares
        '    ' that correspond to a set of initial probabilities
        '    Dim result(1) As Integer
        '    ' set outstanding points assoc. with smaller prob to 0
        '    Dim smallIdx As Integer = 0 ' take a guess
        '    Dim largeIdx As Integer = 1
        '    If (probs(1) < probs(0)) Then ' we were wrong
        '        smallIdx = 1
        '        largeIdx = 0
        '    End If
        '    result(smallIdx) = 0 ' 0 points outstanding
        '    Dim p As Double = probs(largeIdx) ' the larger prob has outstanding points
        '    result(largeIdx) = CInt(Fix(Math.Round(liq * Math.Log(p / (1 - p))))) ' algebra
        '    Return result
        'End Function ' OutstandingFromProbs

        'Static Function SharesEarned(ByVal outstanding() As Integer, ByVal idx As Integer, ByVal shares As Integer, ByVal liq As Double) As Integer
        '    ' shares earned if option idx wins
        '    ' the return includes the points-bet
        '    Dim s As Integer = 0
        '    Dim costOfTrans As Double = 0.0
        '    While (costOfTrans < shares) ' subtle
        '        costOfTrans = costOfTrans(outstanding, idx, s, liq)
        '        s += 1
        '    End While
        '    Return s
        'End Function ' PointsEarned

        Private Shared Function Cost(ByVal outstanding() As Integer, ByVal liq As Double) As Double
            ' cost asociated with a set of outstanding shares
            Dim sum As Double = 0.0
            For i As Integer = 0 To 1
                sum += Math.Exp(outstanding(i) / liq)
            Next i
            Return liq * Math.Log(sum)
        End Function

        Private Shared Function CostOfTrans(ByVal outstanding() As Integer, ByVal idx As Integer, ByVal nShares As Integer, ByVal liq As Double) As Double
            ' cost of a transaction that buys nShares on team idx
            ' given current outstanding points
            Dim after(1) As Integer
            Array.Copy(outstanding, after, 2)
            after(idx) += nShares
            Return Cost(after, liq) - Cost(outstanding, liq)
        End Function

        Private Shared Function CostForOneShare(ByVal outstanding() As Integer, ByVal liq As Double) As Double()
            Dim result(1) As Double
            result(0) = CostOfTrans(outstanding, 0, 1, liq)
            result(1) = CostOfTrans(outstanding, 1, 1, liq)
            Return result
        End Function

        Private Shared Sub ShowVector(ByVal vector() As Double, ByVal dec As Integer, ByVal pre As String)
            For i As Integer = 0 To vector.Length - 1
                Console.Write(pre & vector(i).ToString("F" & dec) & " ")
            Next i
            Console.WriteLine(vbLf)
        End Sub ' ShowVector

        Private Shared Sub ShowVector(ByVal vector() As Integer)
            For i As Integer = 0 To vector.Length - 1
                Console.Write(vector(i) & " ")
            Next i
            Console.WriteLine(vbLf)
        End Sub ' ShowVector

    End Class ' Program class

End Namespace ' ns
